/* -*- mode:c++; tab-width:4; c-basic-offset:4; -*-
 *
 * This file is part of maemo-security-certman
 *
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Juhani Mäkelä <ext-juhani.3.makela@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

/**
 \file aegis_certman.h
 \ingroup libcertman
 \brief The certman library low-level API

  The functions for accessing certificate stores as openSSL's X509
  data structures.

*/

#ifndef AEGIS_CERTMAN_H
#define AEGIS_CERTMAN_H

#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/sha.h>
#include "aegis_common.h"

#ifdef	__cplusplus
extern "C" {
#endif

	/**
	 * \def AEGIS_CERTMAN_DOMAIN_PRIVATE
	 * \brief Private certificate domains can only be accessed by 
	 *        one application, the one that has created them and
	 *        owns them.
	 */
    #define AEGIS_CERTMAN_DOMAIN_PRIVATE 0
	
	/**
	 * \def AEGIS_CERTMAN_DOMAIN_SHARED
	 * \brief Common certificate domain, accessible by all applications
	 */
    #define AEGIS_CERTMAN_DOMAIN_SHARED 1
	
	/**
	 * \typedef domain_handle
	 * \brief A magic cookie reference to a domain opened by \ref
	 * aegis_certman_open_domain and to be used in the domain management 
	 * functions.
	 */
	typedef void* domain_handle;

	/**
	 * \def AEGIS_CERTMAN_DOMAIN_NONE
	 * \brief The value a domain handle cannot have if its properly
	 * opened.
	 */
    #define AEGIS_CERTMAN_DOMAIN_NONE (void*)(0)

	/**
	 * \def AEGIS_KEY_ID_LEN
	 * \brief The length of a key identifier
	 */
    #define AEGIS_KEY_ID_LEN SHA_DIGEST_LENGTH

	/**
	 * \typedef aegis_key_id
	 * \brief A type for a unique key identifier. In practice this is
	 * the SHA1 fingerprint of a (public) key, which is considered 
	 * unique enough.
	 */
	typedef unsigned char aegis_key_id [AEGIS_KEY_ID_LEN];

	/**
	 * \def AEGIS_KEY_ID_STR_LEN
	 * \brief The lenght of a string representing a aegis_key_id
	 * in the printable form, with the terminating NUL included.
	 */
    #define AEGIS_KEY_ID_STR_LEN 2*SHA_DIGEST_LENGTH + 1

    /**
	 * \brief Convert a key id value to string
	 * \param key_id Key id as a byte array
	 * \param to_buf A buffer to hold the default string
	 * \param max_len Size of the buffer, must be >= AEGIS_KEY_ID_STR_LEN
	 * \returns 0 on success, otherwise an error code
	 */
	int aegis_certman_key_id_to_str(aegis_key_id key_id,
									char* to_buf,
									unsigned max_len);

    /**
	 * \brief Convert a string into a key id
	 * \param from_str The string presentation
	 * \param key_id Key id as a byte array
	 * \returns 0 on success, otherwise an error code
	 */
	int aegis_certman_str_to_key_id(char* from_str,
									aegis_key_id key_id);


/// \name General certificate management functions
//@{

    /**
	 * \brief Open a secure certificate store
	 * \param my_cert_store (out) the initial certificate store, 
	 * contains the root X509 certificate from BB5. Should be NULL
	 * when the function is called.
	 * \returns 0 on success, otherwise an error code
	 */
	int aegis_certman_open(X509_STORE** my_cert_store);

	/**
	 * \brief Load and verify the certificates of the given domain. 
	 *        This function can be called multiple times.
	 * \param domain (in) logical name of the domain
	 * \param shared (in) if true, a shared domain is expected, otherwise
	 *                    a private one
	 * \param my_cert_store (in,out) the store where to add the certificates
	 * \return 0 on success, otherwise an error code
	 */
	int aegis_certman_collect(const char* domain, 
							  int shared, 
							  X509_STORE* my_cert_store);

	/**
	 * \brief Close the certificate store and release reserved resources
	 * \param my_cert_store (in) the certificate store to be released
	 * \return 0 on success, otherwise an error code
	 */
	int aegis_certman_close(X509_STORE* my_cert_store);

	//@}

	/// \name Certificate domain management functions
	//@{

	/**
	 * \brief Get a list of existing domains
	 * \param flags (in) the type of domains to iterate (shared, private)
	 * \param cb_func (in) a callback function called once for each
	 *        domain. Parameters: order number, domain name, domain type 
	 *        and userdata pointer.
	 * \param ctx (in) a void pointer passed to the callback function
	 * \return if >= 0, the index where iteration terminated,
	 *         if < 0, an error code
	 * \warning If you modify the domain inside the callback function
	 *        with "add cert" or "rm cert", the results may be quite unpredictable.
	 *        So don't.
	 */
	int aegis_certman_iterate_domains(int flags,
									  aegis_callback* cb_func,
									  void* ctx);

	/**
	 * \brief Set the credentials with which domains are created 
     *        or opened for modification
	 * \param token_name (in) typically a resource token name
	 * \credential * If the domain does not exist already, it is created
	 * with this resource token in the aegis_certman_open_domain
	 * 
	 * \return 0 on success, otherwise an error code
	 */
    int aegis_certman_set_credentials(const char* token_name);

	/**
	 * \brief Open an existing domain or create a new one
	 * \param domain_name (in) logical name of the domain
	 * \param flags (in) type of domain to open/create (see AEGIS_CD_* flags)
	 * \param handle (out) a handle to the domain to be used in subsequent calls
	 * \credential * The resource token of the domain is
	 *   required to make changes in the domain, 
	 *   for instance (aegis-certman-common-ca::CertCASSLAdd)
	 * \return 0 on success, otherwise an error code
	 */
	int aegis_certman_open_domain(const char* domain_name, 
								  int flags, 
								  domain_handle* handle);

	/**
	 * \brief Iterate through all certificates in a domain
	 * \param the_domain (in) a handle to the domain returned by 
	 *        \ref aegis_certman_open_domain
	 * \param cb_func (in) a callback function called once for each
	 *        certificate in the domain. The first parameter
	 *        is the order number of the certificate in the domain
	 *        (starting from 0), the second a pointer to a X509 certificate
	 *        struct and the third is the given ctx pointer.
	 *        If the callback returns other than 0 or -1, the iteration is 
	 *        terminated. If the callback returns 0, the certificate 
	 *        is released right after the callback.
	 * \param ctx (in) a void pointer passed to the callback function
	 * \return if >= 0, the index where iteration terminated,
	 *         if < 0, an error code
	 * \warning If you modify the domain inside the callback function
	 *        with "add cert" or "rm cert", the results may be quite 
	 *        unpredictable. So don't.
	 */
	int aegis_certman_iterate_certs(domain_handle the_domain, 
									int cb_func(int, X509*, void*), 
									void* ctx);


	/**
	 * \brief Load a single certificate identifier by its key fingerprint
	 * \param the_domain (in) a handle to the domain
	 * \param with_id (in) the fingerprint of the public key
	 * \param cert (out) the certificate, if found
	 * \return 0 on success, otherwise an error code. ENOENT
	 * if there is no certificate in the store with the given
	 * key fingerprint.
	 */
	int aegis_certman_load_cert(domain_handle the_domain, 
								aegis_key_id with_id, 
								X509** cert);

	/**
	 * \brief Add a certificate into a domain
	 * \param to_domain (in) a handle to the domain
	 * \param cert (in) the certificate to be added
	 * \return 0 on success, otherwise an error code. EACCES
	 *   if the application does not have the power to modify
	 *   the domain.
	 * \credential * The resource token of the domain is
	 *   required to make changes in the domain
	 *   for instance (aegis-certman-common-ca::CertCASSLAdd)
	 */
	int aegis_certman_add_cert(domain_handle to_domain, X509* cert);

	/**
	 * \brief Add one or more certificates into a domain
	 * \param to_domain (in) a handle to the domain
	 * \param cert_files (in) names of files from which the certificate 
	 *        are to be added
	 * \param count (in) how many file names there are in the list. A NULL also
	 *        terminates the list
	 * \return The number of certificates successfully added or -1.
	 * \credential * The resource token of the domain is
	 *   required to make changes in the domain
	 *   for instance (aegis-certman-common-ca::CertCASSLAdd)
	 */
	int aegis_certman_add_certs(domain_handle to_domain, 
								char* cert_files[], 
								unsigned count);

	/**
	 * \brief Remove a certificate from the domain
	 * \param from_domain (in) 
	 * \param key_id (in) The public key id of the certificate
	 * \return 0 on success, otherwise an error code. EACCESS
	 * if the application does not have the power to modify
	 * the domain.
	 * \credential * The resource token of the domain is
	 *   required to make changes in the domain
	 *   for instance (aegis-certman-common-ca::CertCASSLAdd)
	 */
	int aegis_certman_rm_cert(domain_handle from_domain, aegis_key_id key_id);

	/**
	 * \brief Return the number of certificates in a domain
	 * \param in_domain (in) a handle to the domain
	 * \return >= 0 on success, otherwise -1
	 */
	int aegis_certman_nbrof_certs(domain_handle in_domain);

	/**
	 * \brief Close a domain
	 * \param handle (in) a handle to the domain
	 * \return 0 on success, otherwise and error code
	 * 
	 * Upon closing the changes are updated on the disk, if the
	 * application has permissions to do that.
	 */
	int aegis_certman_close_domain(domain_handle handle);

	/**
	 * \brief Check if the domain has changed
	 * \param handle (in) a handle to the domain
	 * \return 1 if the domain was modified by another process
	 * otherwise return 0
	 */
	int aegis_certman_domain_changed(domain_handle the_domain);

	/**
	 * \brief Get the public key fingerprint out of a certificate
	 * \param of_cert (in) The certificate
	 * \param to_this (out) The fingerprint of the public key
	 * \return 0 on success. EINVAL if the given certificate is 
	 * NULL or otherwise invalid.
	 */
	int aegis_certman_get_key_id(X509* of_cert, aegis_key_id to_this);

	/**
	 * \brief Store a private key into the global private key store.
	 * \param with_id (in) the fingerprint of the corresponding public key
	 * \param the_key (in) the key to be stored
	 * \param with_passwd (in) the password for the container.
	 * The private key is stored in an encrypted PKCS#8 container.
	 * \return 0 on success, an error code otherwise.
	 */
	int aegis_certman_store_key(aegis_key_id with_id, 
								EVP_PKEY* the_key, 
								char* with_passwd);

	/**
	 * \brief Retrieve a private key from the global private key store.
	 * \param with_id (in) the fingerprint of the corresponding public key
	 * \param the_key (out) the private key
	 * \param with_passwd (in) the password for the container.
	 * \return 0 on success, an error code otherwise.
	 * \credential * The resource token of the related
	 *   user-domain is required for reading the key 
	 *   (for instance aegis-certman-common-ca::CertUserSSLUse)
	 */
	int aegis_certman_retrieve_key(aegis_key_id with_id, 
								   EVP_PKEY** the_key, 
								   char* with_passwd);

	/**
	 * \brief Iterate through all of the private keys in the global store
	 * \param cb_func (in) a callback function. The first parameter
	 *        is the order number of the key (starting from 0), 
	 *        the second a pointer to a aegis_key_id 
	 *        struct and the third is the given ctx pointer.
	 * \param ctx (in) a void pointer passed to the callback function
	 * \return if >= 0, the index where iteration terminated,
	 *         if < 0, an error code
	 */
	int aegis_certman_iterate_keys(aegis_callback* cb_func, void* ctx);

	/**
	 * \brief Get a short name composed of the Subject DN name suitable for display
	 * \param of_cert (in) a certificate
	 * \param to_buf (out) the buffer where the nickname is copied to
	 * \param buf_len (in) the maximum length of the name. If the nickname
	 *                     would exceed the length, it's truncated.
	 * \return 0 on success, otherwise an error code.
	 */
	int aegis_certman_get_nickname(X509* of_cert, 
								   char* to_buf, 
								   unsigned buf_len);

	/*
	 * TODO: Document
	 */
	int aegis_certman_has_private_key(X509* cert);
	int aegis_certman_has_private_key_by_id(aegis_key_id key_id);

    /**
     * \brief Verify a certificate using the certman GUI
     * \param cert (in) X509 certificate, usually a server certificate
     * \param timeout (in) Wait for the answer this many seconds
     * \result 0 if the certificate is approved, an error code otherwise
     */
    int aegis_certman_gui_check_certificate(X509 *cert, long timeout);

    /**
     * \brief Get a private key
     * \param with_id (in) The certman_key_id of the related user certificate
     * \param with_credentials (in) The resource token needed to read the key
     * \param timeout (in) Wait for the answer this many seconds
     * \param the_key (out) The private key
     * \result 0 if the key could be retrieved, and error code otherwise
	 * \credential * The resource token of the related
	 *   user-domain is required for reading the key 
	 *   (for instance aegis-certman-common-ca::CertUserSSLUse)
     */
    int aegis_certman_gui_get_private_key(aegis_key_id with_id, 
                                          const char* with_credentials, 
                                          long timeout, 
                                          EVP_PKEY** the_key);

    /**
     * \brief Request a private key and report when it is ready
     * \param with_id (in) The certman_key_id of the related user certificate
     * \param with_credentials (in) The resource token needed to read the key
     * \param timeout (in) Wait for the answer this many seconds
     * \param notify (in) Notification function
     * \param state  (in) A state object to be passed to the notify function
     * \result 0 if the key has been requested, and error code otherwise
     *
     * This function acts just like aegis_certman_gui_get_private_key(), but
     * returns immediately. The requested key will be reported to the caller
     * when it is retrieved, which may take some time. However, if it is not
     * retrieved for timeout seconds, the error is reported to the caller.
     *
     * When the key is retrieved (or when a timeout expires), the notification
     * function supplied via notify parameter will be called. This function
     * takes 3 arguments: state argument of this function, error code, and the
     * retrieved key. The state argument is passed unchanged as supplied by
     * caller and is ignored by this function. It may be any value,
     * including NULL. The error code is 0 if the key was retrieved
     * successfully. It is errno-based.
     *
     * \note If you use this function, you need to call
     * aegis_certman_gui_poll() periodically to look at the bus for
     * requested private key.
     *
     * \see aegis_certman_gui_poll()
     */
    int aegis_certman_gui_get_private_key_async(aegis_key_id with_id,
                                      const char* with_credentials,
                                      long timeout,
                                      void (*notify)(void *, int, EVP_PKEY *),
                                      void* state);

    /**
     * \brief Verify a certificate using certman GUI and report when checked
     * \param cert (in) X509 certificate, usually a server certificate
     * \param timeout (in) Wait for the answer this many seconds
     * \param notify (in) Notification function
     * \param state  (in) A state object to be passed to the notify function
     * \result 0 if requested, and error code otherwise
     *
     * This function acts just like aegis_certman_gui_check_certificate(), but
     * returns immediately. The result will be reported to the caller when the
     * certificate is checked, which may take some time. However, if it is not
     * retrieved for timeout seconds, the error is reported to the caller.
     *
     * When the certificate is checked (or when a timeout expires), the
     * notification function supplied via notify parameter will be
     * called. This function takes 2 arguments: state argument of this
     * function and error code. The state argument is passed unchanged
     * as supplied by caller and is ignored by this function. It may be any
     * value, including NULL. The error code is 0 if the certificate
     * was approved successfully. It is errno-based.
     *
     * \note If you use this function, you need to call
     * aegis_certman_gui_poll() periodically to look at the bus for
     * certificate approval.
     *
     * \see aegis_certman_gui_poll()
     */
    int aegis_certman_gui_check_certificate_async(X509 *cert,
                                      long timeout,
                                      void (*notify)(void*, int),
                                      void* state);

    /**
     * \brief Look for a reply from system bus
     * \param wait (in) Maximum time to wait for incoming reply, milliseconds.
     *                  If <=0, the function returns immediately
     *
     * You need to call this function only if a private key was requested with
     * aegis_certman_gui_get_private_key_async() and before the requested key
     * is retrieved (or a timeout expires). This function blocks for at least
     * wait milliseconds looking for an incoming key on session bus. If the
     * key is found, it is decrypted and reported via the notification function
     * passed to aegis_certman_gui_get_private_key_async().
     *
     * \note If your application uses only one thread, you don't have to worry
     * about this note. This function delivers results of requests in
     * the calling thread. If results are to be delivered in some other thread
     * it must be implemented by the caller. That is, notification functions
     * for pending calls are called in the same thread that has called this
     * function. A notification function may then post reported results
     * to some other thread using the framework API (Qt, GObject).
     *
     * If no operation was started with
     * aegis_certman_gui_get_private_key_async() and this function is called,
     * it will NOT block and will return immediately.
     */
    void aegis_certman_gui_poll(long wait);

//@}

#ifdef	__cplusplus
} // extern "C"
#endif

#endif
